import json
import logging
import ipaddress
from typing import Optional
from datetime import datetime
from contextlib import suppress
from pydantic import BaseModel, validator
from threading import Event as ThreadingEvent

from .helpers import *
from bbot.core.errors import *
from bbot.core.helpers import (
    extract_words,
    split_host_port,
    host_in_host,
    is_domain,
    is_subdomain,
    is_ip,
    domain_stem,
    make_netloc,
    make_ip_type,
    smart_decode,
    get_file_extension,
    validators,
)


log = logging.getLogger("bbot.core.event")


class BaseEvent:

    # Exclude from output modules
    _omit = False
    # Priority, 1-5, lower numbers == higher priority
    _priority = 3
    # Disables certain data validations
    _dummy = False
    # Data validation, if data is a dictionary
    _data_validator = None

    def __init__(
        self,
        data,
        event_type=None,
        source=None,
        module=None,
        scan=None,
        tags=None,
        confidence=100,
        _dummy=False,
        _internal=None,
    ):

        self._id = None
        self._hash = None
        self.__host = None
        self._port = None
        self.__words = None
        self._made_internal = False
        # whether to force-send to output modules
        self._force_output = False

        self.timestamp = datetime.utcnow()

        if tags is None:
            tags = set()

        self._data = None
        self.type = event_type
        self.tags = set(tags)
        self.confidence = int(confidence)

        # for creating one-off events without enforcing source requirement
        self._dummy = _dummy
        self._internal = False

        self.module = module
        self.scan = scan
        if (not self.scan) and (not self._dummy):
            raise ValidationError(f"Must specify scan")

        # check type blacklist
        if self.scan is not None:
            omit_event_types = self.scan.config.get("omit_event_types", [])
            if omit_event_types and self.type in omit_event_types:
                self._omit = True

        self._scope_distance = -1

        try:
            self.data = self._sanitize_data(data)
        except Exception as e:
            import traceback

            log.debug(traceback.format_exc())
            raise ValidationError(f'Error sanitizing event data "{data}" for type "{self.type}": {e}')

        if not self.data:
            raise ValidationError(f'Invalid event data "{data}" for type "{self.type}"')

        self._source = None
        self.source_id = None
        self.source = source
        if (not self.source) and (not self._dummy):
            raise ValidationError(f"Must specify event source")

        if not self._dummy:
            self._setup()

        # internal events are not ingested by output modules
        if not self._dummy:
            # removed this second part because it was making certain sslcert events internal
            if _internal:  # or source._internal:
                self.make_internal()

        self._resolved = ThreadingEvent()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._hash = None
        self._id = None
        self.__host = None
        self._port = None
        self._data = data

    @property
    def host(self):
        """
        An abbreviated representation of the data that allows comparison with other events.
        For host types, this is a hostname.
        This allows comparison of an email or a URL with a domain, and vice versa
            bob@evilcorp.com        --> evilcorp.com
            https://evilcorp.com    --> evilcorp.com
            evilcorp.com:80         --> evilcorp.com

        For IP_* types, this is an instantiated object representing the event's data
        E.g. for IP_ADDRESS, it could be an ipaddress.IPv4Address() or IPv6Address() object
        """
        if self.__host is None:
            self.__host = self._host()
        return self.__host

    @property
    def port(self):
        self.host
        return self._port

    @property
    def host_stem(self):
        """
        An abbreviated representation of hostname that removes the TLD
            E.g. www.evilcorp.com --> www.evilcorp
        """
        if self.host and type(self.host) == str:
            return domain_stem(self.host)
        else:
            return f"{self.host}"

    @property
    def words(self):
        if self.__words is None:
            self.__words = set(self._words())
        return self.__words

    def _words(self):
        return set()

    @property
    def id(self):
        if self._id is None:
            self._id = make_event_id(self.data_id, self.type)
        return self._id

    @property
    def scope_distance(self):
        return self._scope_distance

    @scope_distance.setter
    def scope_distance(self, scope_distance):
        if scope_distance >= 0:
            new_scope_distance = None
            # ensure scope distance does not increase (only allow setting to smaller values)
            if self.scope_distance == -1:
                new_scope_distance = scope_distance
            else:
                new_scope_distance = min(self.scope_distance, scope_distance)
            self._scope_distance = new_scope_distance
            for t in list(self.tags):
                if t.startswith("distance-"):
                    self.tags.remove(t)
            self.tags.add(f"distance-{new_scope_distance}")

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if is_event(source):
            self._source = source
            if source.scope_distance >= 0 and source != self:
                new_scope_distance = int(source.scope_distance)
                # only increment the scope distance if the host changes
                if not self.host == source.host:
                    new_scope_distance += 1
                self.scope_distance = new_scope_distance
            self.source_id = str(source.id)
        elif not self._dummy:
            log.warning(f"Tried to set invalid source on {self}: (got: {source})")

    def get_source(self):
        """
        Takes into account events with the _omit flag
        """
        if getattr(self.source, "_omit", False):
            return self.source.get_source()
        return self.source

    def make_internal(self):
        if not self._made_internal:
            self._internal = True
            self.tags.add("internal")
            self._made_internal = True

    def unmake_internal(self, set_scope_distance=None, force_output=False, emit_trail=True):
        source_trail = []
        if self._made_internal:
            if set_scope_distance is not None:
                self.scope_distance = set_scope_distance
            self._internal = False
            self.tags.remove("internal")
            if force_output:
                self._force_output = True
            self._made_internal = False

        if getattr(self.source, "_internal", False):
            source_scope_distance = None
            if set_scope_distance is not None:
                source_scope_distance = set_scope_distance + 1
            source_trail += self.source.unmake_internal(
                set_scope_distance=source_scope_distance, force_output=force_output
            )
            source_trail.append(self.source)

        if emit_trail and self.scan:
            for e in source_trail:
                self.scan.manager.emit_event(e, release=False)

        return source_trail

    def make_in_scope(self, set_scope_distance=0):
        source_trail = []
        # keep the event internal if the module requests so, unless it's a DNS_NAME
        if getattr(self.module, "_scope_shepherding", True) or self.type in ("DNS_NAME",):
            source_trail = self.unmake_internal(
                set_scope_distance=set_scope_distance, force_output=True, emit_trail=True
            )
        self.scope_distance = set_scope_distance
        if set_scope_distance == 0:
            self.tags.add("in-scope")
        return source_trail

    def _host(self):
        return ""

    def _sanitize_data(self, data):
        if self._data_validator is not None:
            if not isinstance(data, dict):
                raise ValidationError(f"data is not of type dict: {data}")
            data = self._data_validator(**data).dict()
        return self.sanitize_data(data)

    def sanitize_data(self, data):
        return data

    @property
    def data_human(self):
        return self._data_human()

    def _data_human(self):
        return str(self.data)

    @property
    def data_id(self):
        return self._data_id()

    def _data_id(self):
        return self.data

    @property
    def data_graph(self):
        return self._data_graph()

    def _data_graph(self):
        if type(self.data) in (list, dict):
            with suppress(Exception):
                return json.dumps(self.data, sort_keys=True)
        return smart_decode(self.data)

    def _setup(self):
        """
        Perform optional setup, e.g. adding custom tags
        """

    def __contains__(self, other):
        """
        Allows events to be compared using the "in" operator:
        E.g.:
            if some_event in other_event:
                ...
        """
        try:
            other = make_event(other, dummy=True)
        except ValidationError:
            return False
        # if hashes match
        if other == self:
            return True
        # if hosts match
        if self.host and other.host:
            if self.host == other.host:
                return True
            # hostnames and IPs
            return host_in_host(other.host, self.host)
        return False

    def json(self, mode="graph"):
        j = dict()
        for i in ("type", "id", "web_spider_distance"):
            v = getattr(self, i, "")
            if v:
                j.update({i: v})
        data_attr = getattr(self, f"data_{mode}", None)
        if data_attr is not None:
            j["data"] = data_attr
        else:
            j["data"] = smart_decode(self.data)
        j["scope_distance"] = self.scope_distance
        j["scan"] = self.scan.id
        j["timestamp"] = self.timestamp.timestamp()
        source = self.get_source()
        source_id = getattr(source, "id", "")
        if source_id:
            j["source"] = source_id
        if self.tags:
            j.update({"tags": list(self.tags)})
        if self.module:
            j.update({"module": str(self.module)})

        # normalize non-primitive python objects
        for k, v in list(j.items()):
            if k == "data":
                continue
            if type(v) not in (str, int, float, bool, list, type(None)):
                try:
                    j[k] = json.dumps(v, sort_keys=True)
                except Exception:
                    j[k] = smart_decode(v)
        return j

    @property
    def priority(self):
        self_priority = int(max(1, min(5, self._priority)))
        mod_priority = int(max(1, min(5, getattr(self.module, "priority", 1))))
        timestamp = self.timestamp.timestamp()
        return self_priority + mod_priority + (1 / timestamp)

    def __iter__(self):
        """
        For dict(event)
        """
        yield from self.json().items()

    def __lt__(self, other):
        """
        For queue sorting
        """
        return self.priority < int(getattr(other, "priority", 5))

    def __gt__(self, other):
        """
        For queue sorting
        """
        return self.priority > int(getattr(other, "priority", 5))

    def __eq__(self, other):
        try:
            other = make_event(other, dummy=True)
        except ValidationError:
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.id)
        return self._hash

    def __str__(self):
        d = str(self.data)
        return f'{self.type}("{d[:50]}{("..." if len(d) > 50 else "")}", module={self.module}, tags={self.tags})'

    def __repr__(self):
        return str(self)


class DefaultEvent(BaseEvent):
    def sanitize_data(self, data):
        return data


class DictEvent(BaseEvent):
    def _data_human(self):
        return json.dumps(self.data, sort_keys=True)


class DictHostEvent(DictEvent):
    def _host(self):
        return make_ip_type(self.data["host"])


class CODE_REPOSITORY(DictHostEvent):
    class _data_validator(BaseModel):
        url: str
        _validate_url = validator("url", allow_reuse=True)(validators.validate_url)

    def _host(self):
        self.parsed = validators.validate_url_parsed(self.data["url"])
        return make_ip_type(self.parsed.hostname)

    def _data_graph(self):
        return self.data["url"]


class IP_ADDRESS(BaseEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ip = ipaddress.ip_address(self.data)
        self.tags.add(f"ipv{ip.version}")
        if ip.is_private:
            self.tags.add("private")

    def sanitize_data(self, data):
        return validators.validate_host(data)

    def _host(self):
        return ipaddress.ip_address(self.data)


class IP_RANGE(BaseEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        net = ipaddress.ip_network(self.data, strict=False)
        self.tags.add(f"ipv{net.version}")

    def sanitize_data(self, data):
        return str(ipaddress.ip_network(str(data), strict=False))

    def _host(self):
        return ipaddress.ip_network(self.data)


class DNS_NAME(BaseEvent):
    _priority = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if is_subdomain(self.data):
            self.tags.add("subdomain")
        elif is_domain(self.data):
            self.tags.add("domain")

    def sanitize_data(self, data):
        return validators.validate_host(data)

    def _host(self):
        return self.data

    def _words(self):
        stem = self.host_stem
        if "wildcard" in self.tags:
            stem = "".join(stem.split(".")[1:])
        return extract_words(self.host_stem)


class OPEN_TCP_PORT(BaseEvent):
    def sanitize_data(self, data):
        return validators.validate_open_port(data)

    def _host(self):
        host, self._port = split_host_port(self.data)
        return host

    def _words(self):
        if not is_ip(self.host):
            return extract_words(self.host_stem)
        return set()


class URL_UNVERIFIED(BaseEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.web_spider_distance = getattr(self.source, "web_spider_distance", 0)

    def sanitize_data(self, data):
        self.parsed = validators.validate_url_parsed(data)

        # tag as dir or endpoint
        if str(self.parsed.path).endswith("/"):
            self.tags.add("dir")
        else:
            self.tags.add("endpoint")

        parsed_path_lower = str(self.parsed.path).lower()

        url_extension_blacklist = []
        url_extension_httpx_only = []
        scan = getattr(self, "scan", None)
        if scan is not None:
            url_extension_blacklist = [e.lower() for e in scan.config.get("url_extension_blacklist", [])]
            url_extension_httpx_only = [e.lower() for e in scan.config.get("url_extension_httpx_only", [])]

        extension = get_file_extension(parsed_path_lower)
        if extension:
            self.tags.add(f"extension-{extension}")
            if extension in url_extension_blacklist:
                self.tags.add("blacklisted")
            if extension in url_extension_httpx_only:
                self.tags.add("httpx-only")
                self._omit = True

        data = self.parsed.geturl()
        return data

    def with_port(self):
        netloc_with_port = make_netloc(self.host, self.port)
        return self.parsed._replace(netloc=netloc_with_port)

    def _words(self):
        first_elem = self.parsed.path.lstrip("/").split("/")[0]
        if not "." in first_elem:
            return extract_words(first_elem)
        return set()

    def _host(self):
        return make_ip_type(self.parsed.hostname)

    @property
    def port(self):
        if self.parsed.port is not None:
            return self.parsed.port
        elif self.parsed.scheme == "https":
            return 443
        elif self.parsed.scheme == "http":
            return 80


class URL(URL_UNVERIFIED):
    def sanitize_data(self, data):
        if not self._dummy and not any(t.startswith("status-") for t in self.tags):
            raise ValidationError(
                'Must specify HTTP status tag for URL event, e.g. "status-200". Use URL_UNVERIFIED if the URL is unvisited.'
            )
        return super().sanitize_data(data)


class URL_HINT(URL_UNVERIFIED):
    pass


class EMAIL_ADDRESS(BaseEvent):
    def sanitize_data(self, data):
        return validators.validate_email(data)

    def _host(self):
        data = str(self.data).split("@")[-1]
        host, self._port = split_host_port(data)
        return host

    def _words(self):
        return extract_words(self.host_stem)


class HTTP_RESPONSE(URL_UNVERIFIED, DictEvent):
    _priority = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.web_spider_distance = getattr(self.source, "web_spider_distance", 0)
        if not str(self.data.get("status-code", 0)).startswith("3"):
            self.web_spider_distance += 1

    def sanitize_data(self, data):
        url = data.get("url", "")
        self.parsed = validators.validate_url_parsed(url)

        header_dict = {}
        for i in data.get("response-header", "").splitlines():
            if len(i) > 0 and ":" in i:
                k, v = i.split(":", 1)
                k = k.strip().lower()
                v = v.lstrip()
                header_dict[k] = v
        data["header-dict"] = header_dict
        return data

    def _words(self):
        return set()


class VULNERABILITY(DictHostEvent):
    _priority = 1

    def _sanitize_data(self, data):
        data = super()._sanitize_data(data)
        self.tags.add(data["severity"].lower())
        return data

    class _data_validator(BaseModel):
        host: str
        severity: str
        description: str
        url: Optional[str]
        _validate_host = validator("host", allow_reuse=True)(validators.validate_host)
        _validate_severity = validator("severity", allow_reuse=True)(validators.validate_severity)

    def _data_graph(self):
        return f'[{self.data["severity"]}] {self.data["description"]}'


class FINDING(DictHostEvent):
    _priority = 1

    class _data_validator(BaseModel):
        host: str
        description: str
        url: Optional[str]
        _validate_host = validator("host", allow_reuse=True)(validators.validate_host)

    def _data_graph(self):
        return self.data["description"]


class TECHNOLOGY(DictHostEvent):
    _priority = 2

    class _data_validator(BaseModel):
        host: str
        technology: str
        url: Optional[str]
        _validate_host = validator("host", allow_reuse=True)(validators.validate_host)

    def _data_graph(self):
        return self.data["technology"]


class VHOST(DictHostEvent):
    class _data_validator(BaseModel):
        host: str
        vhost: str
        url: Optional[str]
        _validate_host = validator("host", allow_reuse=True)(validators.validate_host)

    def _data_graph(self):
        return self.data["vhost"]


class PROTOCOL(DictHostEvent):
    class _data_validator(BaseModel):
        host: str
        protocol: str
        _validate_host = validator("host", allow_reuse=True)(validators.validate_open_port)

    def _host(self):
        host, self._port = split_host_port(self.data["host"])
        return host

    def _data_graph(self):
        return self.data["protocol"]


def make_event(
    data, event_type=None, source=None, module=None, scan=None, tags=None, confidence=100, dummy=False, internal=None
):
    """
    If data is already an event, simply return it
    """

    if is_event(data):
        if scan is not None and not data.scan:
            data.scan = scan
        if module is not None:
            data.module = module
        if source is not None:
            data.set_source(source)
        if internal == True and not data._made_internal:
            data.make_internal()
        event_type = data.type
        return data
    else:
        if event_type is None:
            event_type = get_event_type(data)
            if not dummy:
                log.debug(f'Autodetected event type "{event_type}" based on data: "{data}"')
        if event_type is None:
            raise ValidationError(f'Unable to autodetect event type from "{data}"')

        event_type = str(event_type).strip().upper()

        # Catch these common whoopsies

        # DNS_NAME <--> IP_ADDRESS confusion
        if event_type in ("DNS_NAME", "IP_ADDRESS"):
            try:
                data = validators.validate_host(data)
            except Exception as e:
                raise ValidationError(f'Error sanitizing event data "{data}" for type "{event_type}": {e}')
            data_is_ip = is_ip(data)
            if event_type == "DNS_NAME" and data_is_ip:
                event_type = "IP_ADDRESS"
            elif event_type == "IP_ADDRESS" and not data_is_ip:
                event_type = "DNS_NAME"

        # DNS_NAME <--> EMAIL_ADDRESS confusion
        if event_type in ("DNS_NAME", "EMAIL_ADDRESS"):
            data_is_email = validators.soft_validate(data, "email")
            if event_type == "DNS_NAME" and data_is_email:
                event_type = "EMAIL_ADDRESS"
            elif event_type == "EMAIL_ADDRESS" and not data_is_email:
                event_type = "DNS_NAME"

        event_class = globals().get(event_type, DefaultEvent)

        return event_class(
            data,
            event_type=event_type,
            source=source,
            module=module,
            scan=scan,
            tags=tags,
            confidence=confidence,
            _dummy=dummy,
            _internal=internal,
        )


def is_event(e):
    return BaseEvent in e.__class__.__mro__



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bbot',
 'bbot.agent',
 'bbot.core',
 'bbot.core.configurator',
 'bbot.core.event',
 'bbot.core.helpers',
 'bbot.core.helpers.depsinstaller',
 'bbot.core.logger',
 'bbot.db',
 'bbot.modules',
 'bbot.modules.deadly',
 'bbot.modules.internal',
 'bbot.modules.output',
 'bbot.modules.report',
 'bbot.scanner',
 'bbot.test']

package_data = \
{'': ['*'], 'bbot': ['wordlists/*']}

install_requires = \
['ansible-runner>=2.2.0,<3.0.0',
 'ansible>=5.7.1,<6.0.0',
 'deepdiff>=5.8.1,<6.0.0',
 'dnspython>=2.2.1,<3.0.0',
 'omegaconf>=2.1.1,<3.0.0',
 'pip>=22.2.2,<23.0.0',
 'psutil>=5.9.0,<6.0.0',
 'pycryptodome>=3.15.0,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests-cache>=0.9.3,<0.10.0',
 'tabulate>=0.8.10,<0.9.0',
 'tldextract>=3.2.0,<4.0.0',
 'websocket-client>=1.3.2,<2.0.0',
 'wordninja>=2.0.0,<3.0.0',
 'xmltojson>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['bbot = bbot.cli:main']}

setup_kwargs = {
    'name': 'bbot',
    'version': '1.0.3.521',
    'description': 'OSINT automation for hackers.',
    'long_description': '![bbot_banner](https://user-images.githubusercontent.com/20261699/158000235-6c1ace81-a267-4f8e-90a1-f4c16884ebac.png)\n\n# BEE·bot\n### OSINT automation for hackers.\n\n[![Python Version](https://img.shields.io/badge/python-3.9+-FF8400)](https://www.python.org) [![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) [![License](https://img.shields.io/badge/license-GPLv3-FF8400.svg)](https://github.com/blacklanternsecurity/bbot/blob/dev/LICENSE) [![Tests](https://github.com/blacklanternsecurity/bbot/workflows/tests/badge.svg)](https://github.com/blacklanternsecurity/bbot/actions?query=workflow%3A"tests") [![Codecov](https://codecov.io/gh/blacklanternsecurity/bbot/branch/dev/graph/badge.svg?token=IR5AZBDM5K)](https://codecov.io/gh/blacklanternsecurity/bbot)\n\n![subdomain demo](https://user-images.githubusercontent.com/20261699/182274919-d4f5aa69-993a-40aa-95d5-f5e69e96026c.gif)\n\n### **BBOT** is a **recursive**, **modular** OSINT framework written in Python.\n\nIt is capable of executing the entire OSINT process in a single command, including subdomain enumeration, port scanning, web screenshots (with its `gowitness` module), vulnerability scanning (with `nuclei`), and much more.\n\nBBOT currently has over **50 modules** and counting.\n\n## Installation\n~~~bash\npipx install bbot\n~~~\nPrerequisites: \n- Python 3.9 or newer MUST be installed\n- `pipx` is recommended as an alternative to `pip` because it installs BBOT in its own Python environment.\n\nIf you need help with installation, please refer to the [wiki](https://github.com/blacklanternsecurity/bbot/wiki#installation).\n\n## Scanning with BBOT\n\n#### Note: the `httpx` module is recommended in most scans because it is used by BBOT to visit webpages. For details, see the [wiki](https://github.com/blacklanternsecurity/bbot/wiki#note-on-the-httpx-module).\n\n### Examples\n~~~bash\n# list modules\nbbot -l\n\n# subdomain enumeration\nbbot --flags subdomain-enum --modules naabu httpx --targets evilcorp.com\n\n# passive modules only\nbbot --flags passive --targets evilcorp.com\n\n# web screenshots with gowitness\nbbot -m naabu httpx gowitness --name my_scan --output-dir . -t evilcorp.com 1.2.3.4/28 4.3.2.1 targets.txt\n\n# web spider (search for emails, etc.)\nbbot -m httpx -c web_spider_distance=2 web_spider_depth=2 -t www.evilcorp.com\n~~~\n\n## Using BBOT as a Python library\n~~~python\nfrom bbot.scanner import Scanner\n\nscan = Scanner("evilcorp.com", "1.2.3.0/24", modules=["naabu"], output_modules=["json"])\nscan.start()\n~~~\n\n# Output\nBBOT outputs to STDOUT by default, but it can output in multiple formats simultaneously (with `--output-module`).\n~~~bash\n# tee to a file\nbbot -f subdomain-enum -t evilcorp.com | tee evilcorp.txt\n\n# output to JSON\nbbot --output-module json -f subdomain-enum -t evilcorp.com | jq\n\n# output to CSV, TXT, and JSON, in current directory\nbbot -o . --output-module human csv json -f subdomain-enum -t evilcorp.com\n~~~\nFor every scan, BBOT generates a unique and mildly-entertaining name like `fuzzy_gandalf`. Output for that scan, including the word cloud and any gowitness screenshots, etc., are saved to a folder by that name in `~/.bbot/scans`. The most recent 20 scans are kept, and older ones are removed. You can change the location of BBOT\'s output with `--output`, and you can also pick a custom scan name with `--name`.\n\nIf you reuse a scan name, it will append to its original output files and leverage the previous word cloud.\n\n# Neo4j\nNeo4j is the funnest (and prettiest) way to view and interact with BBOT data.\n\n![neo4j](https://user-images.githubusercontent.com/20261699/182398274-729f3c48-c23c-4db0-8c2e-8b403c1bf790.png)\n\n- You can get Neo4j up and running with a single docker command:\n~~~bash\ndocker run -p 7687:7687 -p 7474:7474 --env NEO4J_AUTH=neo4j/bbotislife neo4j\n~~~\n- After that, run bbot with `--output-modules neo4j`\n~~~bash\nbbot -f subdomain-enum -t evilcorp.com --output-modules human neo4j\n~~~\n- Browse data at http://localhost:7474\n\n# Usage\n~~~\n$ bbot --help\nusage: bbot [-h] [-t TARGET [TARGET ...]] [-w WHITELIST [WHITELIST ...]] [-b BLACKLIST [BLACKLIST ...]] [-s] [-n SCAN_NAME] [-m MODULE [MODULE ...]] [-l] [-em MODULE [MODULE ...]] [-f FLAG [FLAG ...]]\n            [-rf FLAG [FLAG ...]] [-ef FLAG [FLAG ...]] [-om MODULE [MODULE ...]] [-o DIR] [-c [CONFIG ...]] [--allow-deadly] [-v] [-d] [--force] [-y] [--dry-run] [--current-config] [--save-wordcloud FILE]\n            [--load-wordcloud FILE] [--no-deps | --force-deps | --retry-deps | --ignore-failed-deps] [-a]\n\nBighuge BLS OSINT Tool\n\noptions:\n  -h, --help            show this help message and exit\n  -n SCAN_NAME, --name SCAN_NAME\n                        Name of scan (default: random)\n  -m MODULE [MODULE ...], --modules MODULE [MODULE ...]\n                        Modules to enable. Choices: affiliates,asn,aspnet_viewstate,azure_tenant,binaryedge,blind_ssrf,bypass403,c99,censys,certspotter,cookie_brute,crobat,crt,dnscommonsrv,dnsdumpster,dnszonetransfer,emailformat,ffuf,ffuf_shortnames,generic_ssrf,getparam_brute,github,gowitness,hackertarget,header_brute,host_header,httpx,hunt,hunterio,iis_shortnames,ipneighbor,leakix,massdns,naabu,ntlm,nuclei,passivetotal,pgp,securitytrails,shodan_dns,skymem,smuggler,sslcert,sublist3r,telerik,threatminer,urlscan,viewdns,wappalyzer,wayback,zoomeye\n  -l, --list-modules    List available modules.\n  -em MODULE [MODULE ...], --exclude-modules MODULE [MODULE ...]\n                        Exclude these modules.\n  -f FLAG [FLAG ...], --flags FLAG [FLAG ...]\n                        Enable modules by flag. Choices: active,aggressive,brute-force,deadly,passive,portscan,report,safe,slow,subdomain-enum,web\n  -rf FLAG [FLAG ...], --require-flags FLAG [FLAG ...]\n                        Disable modules that don\'t have these flags (e.g. --require-flags passive)\n  -ef FLAG [FLAG ...], --exclude-flags FLAG [FLAG ...]\n                        Disable modules with these flags. (e.g. --exclude-flags brute-force)\n  -om MODULE [MODULE ...], --output-modules MODULE [MODULE ...]\n                        Output module(s). Choices: csv,http,human,json,neo4j,websocket\n  -o DIR, --output-dir DIR\n  -c [CONFIG ...], --config [CONFIG ...]\n                        custom config file, or configuration options in key=value format: \'modules.shodan.api_key=1234\'\n  --allow-deadly        Enable running modules tagged as "deadly"\n  -v, --verbose         Be more verbose\n  -d, --debug           Enable debugging\n  --force               Run scan even if module setups fail\n  -y, --yes             Skip scan confirmation prompt\n  --dry-run             Abort before executing scan\n  --current-config      Show current config in YAML format\n\nTarget:\n  -t TARGET [TARGET ...], --targets TARGET [TARGET ...]\n                        Targets to seed the scan\n  -w WHITELIST [WHITELIST ...], --whitelist WHITELIST [WHITELIST ...]\n                        What\'s considered in-scope (by default it\'s the same as --targets)\n  -b BLACKLIST [BLACKLIST ...], --blacklist BLACKLIST [BLACKLIST ...]\n                        Don\'t touch these things\n  -s, --strict-scope    Don\'t consider subdomains of target/whitelist to be in-scope\n\nWord cloud:\n  Save/load wordlist of common words gathered during a scan\n\n  --save-wordcloud FILE\n                        Output wordcloud to custom file when the scan completes\n  --load-wordcloud FILE\n                        Load wordcloud from a custom file\n\nModule dependencies:\n  Control how modules install their dependencies\n\n  --no-deps             Don\'t install module dependencies\n  --force-deps          Force install all module dependencies\n  --retry-deps          Try again to install failed module dependencies\n  --ignore-failed-deps  Run modules even if they have failed dependencies\n\nAgent:\n  Report back to a central server\n\n  -a, --agent-mode      Start in agent mode\n~~~\n\n# BBOT Config\nBBOT loads its config from these places in the following order:\n\n- `~/.config/bbot/defaults.yml`\n- `~/.config/bbot/bbot.yml` <-- Use this one as your main config\n- `~/.config/bbot/secrets.yml` <-- Use this one for sensitive stuff like API keys\n- command line (via `--config`)\n\nCommand-line arguments take precedence over all others. You can give BBOT a custom config file with `--config myconf.yml`, or individual arguments like this: `--config http_proxy=http://127.0.0.1:8080 modules.shodan_dns.api_key=1234`. To display the full and current BBOT config, including any command-line arguments, use `bbot --current-config`.\n\nFor explanations of config options, see `defaults.yml` or the [wiki](https://github.com/blacklanternsecurity/bbot/wiki#yaml-config)\n\n# Modules\n| Module           | Needs API Key   | Description                                                       | Flags                                              | Produced Events                                      |\n|------------------|-----------------|-------------------------------------------------------------------|----------------------------------------------------|------------------------------------------------------|\n| aspnet_viewstate |                 | Parse web pages for viewstates and check them against blacklist3r | active,safe,web                                    | VULNERABILITY                                        |\n| bypass403        |                 | Check 403 pages for common bypasses                               | active,aggressive,web                              | FINDING                                              |\n| cookie_brute     |                 | Check for common HTTP cookie parameters                           | active,aggressive,brute-force,slow,web             | FINDING                                              |\n| dnszonetransfer  |                 | Attempt DNS zone transfers                                        | active,safe,subdomain-enum                         | DNS_NAME                                             |\n| ffuf             |                 | A fast web fuzzer written in Go                                   | active,aggressive,brute-force,deadly,web           | URL                                                  |\n| ffuf_shortnames  |                 | Use ffuf in combination IIS shortnames                            | active,aggressive,brute-force,web                  | URL                                                  |\n| generic_ssrf     |                 | Check for generic SSRFs                                           | active,aggressive,web                              | VULNERABILITY                                        |\n| getparam_brute   |                 | Check for common HTTP GET parameters                              | active,aggressive,brute-force,slow,web             | FINDING                                              |\n| gowitness        |                 | Take screenshots of webpages                                      | active,safe,web                                    | SCREENSHOT                                           |\n| header_brute     |                 | Check for common HTTP header parameters                           | active,aggressive,brute-force,slow,web             | FINDING                                              |\n| host_header      |                 | Try common HTTP Host header spoofing techniques                   | active,aggressive,web                              | FINDING                                              |\n| httpx            |                 | Visit webpages. Many other modules rely on httpx                  | active,safe,web                                    | HTTP_RESPONSE,URL                                    |\n| hunt             |                 | Watch for commonly-exploitable HTTP parameters                    | active,safe,web                                    | FINDING                                              |\n| iis_shortnames   |                 | Check for IIS shortname vulnerability                             | active,safe                                        | URL_HINT                                             |\n| naabu            |                 | Execute port scans with naabu                                     | active,aggressive,portscan                         | OPEN_TCP_PORT                                        |\n| ntlm             |                 | Watch for HTTP endpoints that support NTLM authentication         | active,safe,web                                    | DNS_NAME,FINDING                                     |\n| nuclei           |                 | Fast and customisable vulnerability scanner                       | active,aggressive,deadly,web                       | VULNERABILITY                                        |\n| smuggler         |                 | Check for HTTP smuggling                                          | active,aggressive,brute-force,slow,web             | FINDING                                              |\n| sslcert          |                 | Visit open ports and retrieve SSL certificates                    | active,email-enum,safe,subdomain-enum              | DNS_NAME,EMAIL_ADDRESS                               |\n| telerik          |                 | Scan for critical Telerik vulnerabilities                         | active,aggressive,web                              | FINDING,VULNERABILITY                                |\n| vhost            |                 | Fuzz for virtual hosts                                            | active,aggressive,brute-force,deadly,slow,web      | DNS_NAME,VHOST                                       |\n| wappalyzer       |                 | Extract technologies from web responses                           | active,safe,web                                    | TECHNOLOGY                                           |\n| affiliates       |                 | Summarize affiliate domains at the end of a scan                  | passive,report,safe                                |                                                      |\n| asn              |                 | Query bgpview.io for ASNs                                         | passive,report,safe,subdomain-enum                 | ASN                                                  |\n| azure_tenant     |                 | Query Azure for tenant sister domains                             | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| binaryedge       | X               | Query the BinaryEdge API                                          | passive,safe,subdomain-enum                        | DNS_NAME,EMAIL_ADDRESS,IP_ADDRESS,OPEN_PORT,PROTOCOL |\n| c99              | X               | Query the C99 API for subdomains                                  | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| censys           | X               | Query the Censys API                                              | email-enum,passive,safe,subdomain-enum             | DNS_NAME,EMAIL_ADDRESS,IP_ADDRESS,OPEN_PORT,PROTOCOL |\n| certspotter      |                 | Query Certspotter\'s API for subdomains                            | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| crobat           |                 | Query Project Crobat for subdomains                               | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| crt              |                 | Query crt.sh (certificate transparency) for subdomains            | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| dnscommonsrv     |                 | Check for common SRV records                                      | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| dnsdumpster      |                 | Query dnsdumpster for subdomains                                  | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| emailformat      |                 | Query email-format.com for email addresses                        | email-enum,passive,safe                            | EMAIL_ADDRESS                                        |\n| github           | X               | Query Github\'s API for related repositories                       | passive,safe,subdomain-enum                        | URL_UNVERIFIED                                       |\n| hackertarget     |                 | Query the hackertarget.com API for subdomains                     | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| hunterio         | X               | Query hunter.io for emails                                        | email-enum,passive,safe,subdomain-enum             | DNS_NAME,EMAIL_ADDRESS,URL_UNVERIFIED                |\n| ipneighbor       |                 | Look beside IPs in their surrounding subnet                       | aggressive,passive,subdomain-enum                  | IP_ADDRESS                                           |\n| leakix           |                 | Query leakix.net for subdomains                                   | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| massdns          |                 | Brute-force subdomains with massdns (highly effective)            | aggressive,brute-force,passive,slow,subdomain-enum | DNS_NAME                                             |\n| passivetotal     | X               | Query the PassiveTotal API for subdomains                         | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| pgp              |                 | Query common PGP servers for email addresses                      | email-enum,passive,safe                            | EMAIL_ADDRESS                                        |\n| securitytrails   | X               | Query the SecurityTrails API for subdomains                       | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| shodan_dns       | X               | Query Shodan for subdomains                                       | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| skymem           |                 | Query skymem.info for email addresses                             | email-enum,passive,safe                            | EMAIL_ADDRESS                                        |\n| sublist3r        |                 | Query sublist3r\'s API for subdomains                              | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| threatminer      |                 | Query threatminer\'s API for subdomains                            | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| urlscan          |                 | Query urlscan.io for subdomains                                   | passive,safe,subdomain-enum                        | DNS_NAME,URL_UNVERIFIED                              |\n| viewdns          |                 | Query viewdns.info\'s reverse whois for related domains            | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n| wayback          |                 | Query archive.org\'s API for subdomains                            | passive,safe,subdomain-enum                        | DNS_NAME,URL_UNVERIFIED                              |\n| zoomeye          | X               | Query ZoomEye\'s API for subdomains                                | passive,safe,subdomain-enum                        | DNS_NAME                                             |\n',
    'author': 'TheTechromancer',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/blacklanternsecurity/bbot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
import threading
from time import sleep

log = logging.getLogger("bbot.core.helpers.threadpool")

from .cache import CacheDict
from ...core.errors import ScanCancelledError


class ThreadPoolWrapper:
    """
    Layers more granular control overtop of a shared thread pool
    Allows setting lower thread limits for modules, etc.
    """

    def __init__(self, executor, max_workers=None):
        self.executor = executor
        self.max_workers = max_workers
        self.futures = set()
        self._future_lock = threading.Lock()
        self._submit_task_lock = threading.Lock()

    def submit_task(self, callback, *args, **kwargs):
        with self._submit_task_lock:
            if self.max_workers is not None:
                while self.num_tasks > self.max_workers:
                    sleep(0.1)
            try:
                future = self.executor.submit(callback, *args, **kwargs)
            except RuntimeError as e:
                raise ScanCancelledError(e)
        with self._future_lock:
            self.futures.add(future)
        return future

    @property
    def num_tasks(self):
        with self._future_lock:
            for f in list(self.futures):
                if f.done():
                    self.futures.remove(f)
            return len(self.futures) + (1 if self._submit_task_lock.locked() else 0)

    def shutdown(self, *args, **kwargs):
        self.executor.shutdown(*args, **kwargs)


def as_completed(fs):
    fs = list(fs)
    while fs:
        result = False
        for i, f in enumerate(fs):
            if f.done():
                result = True
                future = fs.pop(i)
                if future._state in ("CANCELLED", "CANCELLED_AND_NOTIFIED"):
                    continue
                yield future
                break
        if not result:
            sleep(0.05)


class _Lock:
    def __init__(self, name):
        self.name = name
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()


class NamedLock:
    """
    Returns a unique threading.Lock() based on a provided string

    Useful for preventing multiple operations from occuring on the same data in parallel
    E.g. simultaneous DNS lookups on the same hostname
    """

    def __init__(self, max_size=1000):
        self._cache = CacheDict(max_size=max_size)

    def get_lock(self, name):
        try:
            return self._cache.get(name)
        except KeyError:
            new_lock = _Lock(name)
            self._cache.put(name, new_lock)
            return new_lock



import json
import subprocess

from bbot.modules.base import BaseModule


technology_map = {"f5 bigip": "bigip", "microsoft asp.net": "asp"}


class nuclei(BaseModule):

    watched_events = ["URL", "TECHNOLOGY"]
    produced_events = ["VULNERABILITY"]
    flags = ["active", "aggressive", "web"]
    meta = {"description": "Fast and customisable vulnerability scanner"}

    batch_size = 100
    options = {
        "version": "2.7.3",
        "tags": "",
        "templates": "",
        "severity": "",
        "ratelimit": 150,
        "concurrency": 25,
        "mode": "severe",
        "etags": "intrusive",
    }
    options_desc = {
        "version": "nuclei version",
        "tags": "execute a subset of templates that contain the provided tags",
        "templates": "template or template directory paths to include in the scan",
        "severity": "Filter based on severity field available in the template.",
        "ratelimit": "maximum number of requests to send per second (default 150)",
        "concurrency": "maximum number of templates to be executed in parallel (default 25)",
        "mode": "technology | severe | manual. Technology: Only activate based on technology events that match nuclei tags. On by default. Severe: Only critical and high severity templates without intrusive. Manual: Fully manual settings",
        "etags": "tags to exclude from the scan",
    }
    deps_ansible = [
        {
            "name": "Download nuclei",
            "unarchive": {
                "src": "https://github.com/projectdiscovery/nuclei/releases/download/v{BBOT_MODULES_NUCLEI_VERSION}/nuclei_{BBOT_MODULES_NUCLEI_VERSION}_linux_amd64.zip",
                "include": "nuclei",
                "dest": "{BBOT_TOOLS}",
                "remote_src": True,
            },
        }
    ]
    in_scope_only = True

    def setup(self):

        self.templates = self.config.get("templates")
        self.tags = self.config.get("tags")
        self.etags = self.config.get("etags")
        self.severity = self.config.get("severity")
        self.iserver = self.scan.config.get("interactsh_server", None)
        self.itoken = self.scan.config.get("interactsh_token", None)

        self.template_stats = self.helpers.download(
            "https://raw.githubusercontent.com/projectdiscovery/nuclei-templates/master/TEMPLATES-STATS.json",
            cache_hrs=72,
        )
        if not self.template_stats:
            self.warning(f"Failed to download nuclei template stats.")
            if self.config.get("mode ") == "technology":
                self.warning("Can't run with technology_mode set to true without template tags JSON")
                return False
        else:
            with open(self.template_stats) as f:
                self.template_stats_json = json.load(f)
                try:
                    self.tag_list = [e.get("name", "") for e in self.template_stats_json.get("tags", [])]
                except Exception as e:
                    self.warning(f"Failed to parse template stats: {e}")
                    return False

        if self.config.get("mode") not in ("technology", "severe", "manual"):
            self.warning(f"Unable to intialize nuclei: invalid mode selected: [{self.config.get('mode')}]")
            return False

        if self.config.get("mode") == "technology":
            self.info(
                "Running nuclei in TECHNOLOGY mode. Scans will only be performed against detected TECHNOLOGY events that match nuclei template tags"
            )
            if "wappalyzer" not in self.scan.modules:
                self.hugewarning(
                    "You are running nuclei in technology mode without wappalyzer to emit technologies. It will never execute unless another module is issuing technologies"
                )

        if self.config.get("mode") == "severe":
            self.info(
                "Running nuclei in SEVERE mode. Only critical and high severity templates will be used. Tag setting will be IGNORED."
            )
            self.severity = "critical,high"
            self.tags = ""

        if self.config.get("mode") == "manual":
            self.info(
                "Running nuclei in MANUAL mode. Settings will be passed directly into nuclei with no modification"
            )
        return True

    def handle_batch(self, *events):

        if self.config.get("mode") == "technology":

            tags_to_scan = {}
            for e in events:
                if e.type == "TECHNOLOGY":
                    reported_tag = e.data.get("technology", "")
                    if reported_tag in technology_map.keys():
                        reported_tag = technology_map[reported_tag]
                    if reported_tag in self.tag_list:
                        tag = e.data.get("technology", "")
                        if tag not in tags_to_scan.keys():
                            tags_to_scan[tag] = [e]
                        else:
                            tags_to_scan[tag].append(e)

            self.debug(f"finished processing this batch's tags with {str(len(tags_to_scan.keys()))} total tags")

            for t in tags_to_scan.keys():
                nuclei_input = [e.data["url"] for e in tags_to_scan[t]]
                taglist = self.tags.split(",")
                taglist.append(t)
                override_tags = ",".join(taglist).lstrip(",")
                self.verbose(f"Running nuclei against {str(len(nuclei_input))} host(s) with the {t} tag")
                for severity, template, host, name in self.execute_nuclei(nuclei_input, override_tags=override_tags):
                    source_event = self.correlate_event(events, host)
                    if source_event == None:
                        continue
                    self.emit_event(
                        {
                            "severity": severity,
                            "host": str(source_event.host),
                            "url": host,
                            "description": f"template: {template}, name: {name}",
                        },
                        "VULNERABILITY",
                        source_event,
                    )

        else:
            nuclei_input = [str(e.data) for e in events]
            for severity, template, host, name in self.execute_nuclei(nuclei_input):
                source_event = self.correlate_event(events, host)
                if source_event == None:
                    continue
                self.emit_event(
                    {
                        "severity": severity,
                        "host": str(source_event.host),
                        "url": host,
                        "description": f"template: {template}, name: {name}",
                    },
                    "VULNERABILITY",
                    source_event,
                )

    def correlate_event(self, events, host):
        for event in events:
            if host in event:
                return event
        self.warning("Failed to correlate nuclei result with event")

    def execute_nuclei(self, nuclei_input, override_tags=""):

        command = [
            "nuclei",
            "-silent",
            "-json",
            "-update-directory",
            f"{self.helpers.tools_dir}/nuclei-templates",
            "-rate-limit",
            self.config.get("ratelimit"),
            "-concurrency",
            str(self.config.get("concurrency")),
            # "-r",
            # self.helpers.resolver_file,
        ]

        for cli_option in ("severity", "templates", "iserver", "itoken", "etags"):
            option = getattr(self, cli_option)

            if option:
                command.append(f"-{cli_option}")
                command.append(option)

        if override_tags:
            command.append(f"-tags")
            command.append(override_tags)
        else:
            setup_tags = getattr(self, "tags")
            if setup_tags:
                command.append(f"-tags")
                command.append(setup_tags)

        if self.scan.config.get("interactsh_disable") == True:
            self.info("Disbling interactsh in accordance with global settings")
            command.append("-no-interactsh")

        for line in self.helpers.run_live(command, input=nuclei_input, stderr=subprocess.DEVNULL):
            try:
                j = json.loads(line)
            except json.decoder.JSONDecodeError:
                self.debug(f"Failed to decode line: {line}")
                continue
            template = j.get("template-id", "")

            # try to get the specific matcher name
            name = j.get("matcher-name", "")

            # fall back to regular name
            if not name:
                self.debug(
                    f"Couldn't get matcher-name from nuclei json, falling back to regular name. Template: [{template}]"
                )
                name = j.get("info", {}).get("name", "")

            severity = j.get("info", {}).get("severity", "").upper()
            host = j.get("host", "")

            if template and name and severity and host:
                yield (severity, template, host, name)
            else:
                self.debug("Nuclei result missing one or more required elements, not reporting. JSON: ({j})")

    def cleanup(self):
        resume_file = self.helpers.current_dir / "resume.cfg"
        resume_file.unlink(missing_ok=True)
